/*---------------------------------------------------------------------------*
  Project:  G2D Program Sample
  File:     heiho_red_walk.c

  Copyright 2004 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: $
  $NoKeywords: $
 *---------------------------------------------------------------------------*/

#include <nitro.h>
#include <nnsys/g2d.h>
#include "task_priority.h"
#include "read_key.h"
#include "utility.h"
#include "vram_map.h"
#include "enemy.h"
#include "geography.h"
#include "effect.h"
#include "renderer.h"

// Declaration of internal functions
static void MoveX_WithGroundCheckHeiho(EnemyTask* task);
static void MoveY_WithGroundCheckHeiho(EnemyTask* task);
static inline void SetStarEffect(EnemyTask* yoshi);

/*---------------------------------------------------------------------------*
  Name:         HeihoRedWalk

  Description:  Displays a waling Heiho.
                        
  Arguments:    None
                
  Returns:      None
 *---------------------------------------------------------------------------*/

void HeihoRedWalk(void)
{
    EnemyTask* pSelfTask   = (EnemyTask *)TaskCtrl_GetCurrentTask();
    u16          rotate  = 0;   // Angle of rotation
    NNSG2dFVec2  scale;         // Scale
    NNSG2dFVec2  pos;           // Cell render position
    
    MoveX_WithGroundCheckHeiho(pSelfTask);
    MoveY_WithGroundCheckHeiho(pSelfTask);
       
    scale.x = pSelfTask->size.x;
    scale.y = pSelfTask->size.y;

    if (pSelfTask->direction == RIGHT)
    {
        scale.x *= -1;
    }

    // Calculate the display X coordinate (If the difference between this coordinate and Yoshi is 0, this is the center of the screen)
    pos.x = FX32_SCREEN_CENTER_X + pSelfTask->pos.x - gYoshiPos.x; 
        
    // Calculate the display Y coordinate (FX32_ONE * 14 is the distance from the center of Heiho to its feet.
    pos.y = pSelfTask->pos.y - FX_Mul(FX32_ONE * 14, pSelfTask->size.y) - FX32_ONE * gBgVScroll;

    // Material configuration
    MaterialSetup();
        
    // Update the animation
    NNS_G2dTickCellAnimation( pSelfTask->cellAnim, FX32_ONE );

    // Render
    //
    // Renders the cell using the Renderer module
    //

    // Configure the image data and palette data the renderer references.
    NNS_G2dSetRendererImageProxy( &render3D, &objDataHeihoRed.imageProxy, &objDataHeihoRed.paletteProxy );
    NNS_G2dPushMtx();
        NNS_G2dTranslate( pos.x, pos.y, 0);
        NNS_G2dRotZ( FX_SinIdx( rotate ), FX_CosIdx( rotate ) );
        NNS_G2dScale( scale.x, scale.y, FX32_ONE );
        NNS_G2dDrawCellAnimation( pSelfTask->cellAnim );
    NNS_G2dPopMtx();


    // Deallocate the task if the distance from Yoshi is greater than the constant
    if (FX_ABS(pSelfTask->pos.x - gYoshiPos.x) > FX32_ONE * 512) 
    {
        pSelfTask->common.reqReleaseFlag = TRUE;           
    }

    // If hit points are 0
    if (pSelfTask->hitpoint == 0)
    {
        SetStarEffect(pSelfTask);
        pSelfTask->common.func = HeihoRedDie_Init;     // Change task           
    }
}

/*---------------------------------------------------------------------------*
  Name:         MoveX_WithGroundCheckHeiho

  Description:  Performs Heiho's X-direction movement process.
                        
  Arguments:    None
                
  Returns:      None
 *---------------------------------------------------------------------------*/

static void MoveX_WithGroundCheckHeiho(EnemyTask* task)
{
    fx32 centerHeight;
    fx32 pos_x_backup = task->pos.x;            // Save the current X coordinate
    fx32 diff;
    
    task->pos.x += task->velocity.x;            // X movement

    centerHeight = GetGeographicHeight(task->pos.x);  // Get the field height

    diff = FX_ABS(centerHeight - task->pos.y);  // The difference in elevation of the field and character

    if (diff < FX32_ONE * 2)                    // Match to the height if it is the height difference is less than two
    {                                           // Effect of climbing a slanted board
        task->pos.y = centerHeight;
    }
    else if (centerHeight < task->pos.y)        // Get rid of side movement is the ground is high
    {                                           // Collision with wall effect
        task->pos.x = pos_x_backup;
        task->velocity.x *= -1;                 // Advance in the opposite direction
        task->direction = (task->velocity.x > 0) ? RIGHT : LEFT ;       
    }
    else                                        // Fall if Yoshi is in a higher position than the ground.                                       
    {
        task->velocity.y -= FX32_ONE * 0.2;     // Gravity free-fall
    }
}

/*---------------------------------------------------------------------------*
  Name:         MoveY_WithGroundCheckHeiho

  Description:  Performs Heiho's Y-direction movement process.
                        
  Arguments:    None
                
  Returns:      None
 *---------------------------------------------------------------------------*/

static void MoveY_WithGroundCheckHeiho(EnemyTask* task)
{
    fx32 centerHeight;

    centerHeight = GetGeographicHeight(task->pos.x);  // Get the field height (after finishing the X movement)

    task->pos.y -= task->velocity.y;        // Y movement
    
    if (task->pos.y > centerHeight)         // Land on the ground
    {
        task->velocity.y    = 0;    
        task->pos.y = centerHeight; 
    }       
}

/*---------------------------------------------------------------------------*
  Name:         SetStarEffect

  Description:  Generates a star task (a damage effect when an enemy is stepped on)
                This causes a heavy burden, since it generates 8 animations.
                        
  Arguments:    None
                
  Returns:      None
 *---------------------------------------------------------------------------*/

static inline void SetStarEffect(EnemyTask* heiho)
{    
    EffectTask* task;
    fx32 x = heiho->pos.x - FX32_ONE * 4;
    fx32 y = heiho->pos.y - FX_Mul(FX32_ONE * 16, heiho->size.y);
            
    // 0
    task = (EffectTask*)TaskCtrl_SetTask(StarBirth, TASK_PRIORITY_EFFECT);
    task->pos.x = x;
    task->pos.y = y;            
    task->velocity.x = FX32_ONE * 2;
    task->velocity.y = FX32_ONE * 2;
    // 1
    task = (EffectTask*)TaskCtrl_SetTask(StarBirth, TASK_PRIORITY_EFFECT);
    task->pos.x = x;
    task->pos.y = y;            
    task->velocity.x = FX32_ONE * -2;
    task->velocity.y = FX32_ONE * 2;
    // 2
    task = (EffectTask*)TaskCtrl_SetTask(StarBirth, TASK_PRIORITY_EFFECT);
    task->pos.x = x;
    task->pos.y = y;            
    task->velocity.x = FX32_ONE * 2;
    task->velocity.y = FX32_ONE * -2;
    // 3
    task = (EffectTask*)TaskCtrl_SetTask(StarBirth, TASK_PRIORITY_EFFECT);
    task->pos.x = x;
    task->pos.y = y;            
    task->velocity.x = FX32_ONE * -2;
    task->velocity.y = FX32_ONE * -2;
    // 4
    task = (EffectTask*)TaskCtrl_SetTask(StarBirth, TASK_PRIORITY_EFFECT);
    task->pos.x = x;
    task->pos.y = y;            
    task->velocity.x = FX32_ONE * 3;
    task->velocity.y = FX32_ONE * 0;
    // 5
    task = (EffectTask*)TaskCtrl_SetTask(StarBirth, TASK_PRIORITY_EFFECT);
    task->pos.x = x;
    task->pos.y = y;            
    task->velocity.x = FX32_ONE * 0;
    task->velocity.y = FX32_ONE * 3;
    // 6
    task = (EffectTask*)TaskCtrl_SetTask(StarBirth, TASK_PRIORITY_EFFECT);
    task->pos.x = x;
    task->pos.y = y;            
    task->velocity.x = FX32_ONE * -3;
    task->velocity.y = FX32_ONE * 0;
    // 7
    task = (EffectTask*)TaskCtrl_SetTask(StarBirth, TASK_PRIORITY_EFFECT);
    task->pos.x = x;
    task->pos.y = y;            
    task->velocity.x = FX32_ONE * 0;
    task->velocity.y = FX32_ONE * -3;
}


