/*---------------------------------------------------------------------------*
  Project:  G2D Program Sample
  File:     renderer.c

  Copyright 2004 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: $
  $NoKeywords: $
 *---------------------------------------------------------------------------*/

#include "common.h"
#include "renderer.h"

// Declaration of internal functions
static BOOL CallBackAddOamMain( const GXOamAttr* pOam, u16 affineIndex, BOOL bDoubleAffine );
static BOOL CallBackAddOamSub( const GXOamAttr* pOam, u16 affineIndex, BOOL bDoubleAffine );
static u16 CallBackAddAffineMain( const MtxFx22* mtx );
static u16 CallBackAddAffineSub( const MtxFx22* mtx );
static BOOL CallbackRegistTransferTask( NNS_GFD_DST_TYPE type, u32 dstAddr, void* pSrc, u32 szByte );
static void InitOamManager(NNSG2dOamManagerInstance* pOamManager, NNSG2dOamType type);
static void InitRenderer3D(
    NNSG2dRendererInstance* pRenderer, 
    NNSG2dRenderSurface* pMainSurface);
static void InitRenderer2D(
    NNSG2dRendererInstance* pRenderer2D,
    NNSG2dRenderSurface*    pMainSurface2D,
    NNSG2dRenderSurface*    pSubSurface2D);

// Global variables
NNSG2dRendererInstance      render3D;      /* Renderer for 3D rendering  */
NNSG2dRendererInstance      render2D;      /* Renderer for 2D rendering  */
NNSG2dOamManagerInstance    sMainOamMan;   /* Callback process target OAM manager */
NNSG2dOamManagerInstance    sSubOamMan;    /* Callback process target OAM manager */

// Internal variables
static NNSG2dRenderSurface      mainSurface3D; /* Main screen Surface */
static NNSG2dRenderSurface      mainSurface2D; /* 2D Main screen Surface */
static NNSG2dRenderSurface       subSurface2D; /* 2D Sub screen  Surface */
/* Work region for the VRAM  transmission manager  */
static NNSGfdVramTransferTask   sTransferTaskArray[NUM_OF_VRAMTASK];
/*  Work region for the cell  transmission manager  */
static NNSG2dCellTransferState  sVramStateArray[NUM_OF_VRAMSTATE];

/* OBJ Size Table */
const ObjWH ObjShapeAndObjSizeTable[3][4] =
{
    {{ 8, 8},{16,16},{32,32},{64,64}},
    {{16, 8},{32, 8},{32,16},{64,32}},
    {{ 8,16},{ 8,32},{16,32},{32,64}}
};

/*---------------------------------------------------------------------------*
  Name:         BeginRendering3D

  Description:  Performs pre-Renderer rendering configurations 

  Arguments:    None

  Returns:      None
 *---------------------------------------------------------------------------*/

void BeginRendering3D(void)
{
    NNS_G2dBeginRendering( &render3D );
}

/*---------------------------------------------------------------------------*
  Name:         EndRendering3D

  Description:  Call after Renderer rendering has finished

  Arguments:    None

  Returns:      None
 *---------------------------------------------------------------------------*/

void EndRendering3D(void)
{
    NNS_G2dEndRendering();  
}

/*---------------------------------------------------------------------------*
  Name:         BeginRendering2D

  Description:  Performs pre-Renderer rendering configurations 

  Arguments:    None

  Returns:      None
 *---------------------------------------------------------------------------*/

void BeginRendering2D(void)
{
    NNS_G2dBeginRendering( &render2D );
}

/*---------------------------------------------------------------------------*
  Name:         EndRendering2D

  Description:  Call after Renderer rendering has finished

  Arguments:    None

  Returns:      None
 *---------------------------------------------------------------------------*/

void EndRendering2D(void)
{
    NNS_G2dEndRendering();

    // Writes out the contents of the buffer to HW
    NNS_G2dApplyAndResetOamManagerBuffer( &sMainOamMan );
    NNS_G2dApplyAndResetOamManagerBuffer( &sSubOamMan );        
}

/*---------------------------------------------------------------------------*
  Name:         InitOamManager

  Description:  Initializes the OAM manager entity.

  Arguments:    pOamManager:    The pointer to the OAM manager to initialize.
                type:           Degignates towards which graphics engine the initialization
                                is directed.

  Returns:      None
 *---------------------------------------------------------------------------*/
 
void InitOamManager(NNSG2dOamManagerInstance* pOamManager, NNSG2dOamType type)
{
    BOOL result;
    SDK_NULL_ASSERT( pOamManager );

    result = NNS_G2dGetNewManagerInstance( pOamManager, 0, NUM_OF_OAM - 1, type );
    SDK_ASSERT( result );
    result = NNS_G2dInitManagerInstanceAffine( pOamManager, 0, NUM_OF_AFFINE - 1 );
    SDK_ASSERT( result );
}

/*---------------------------------------------------------------------------*
  Name:         CallBackCulling

  Description:  This is a function called to determine whether the cell is displayed
                inside the NNS_G2Draw* function.
                Determines whether or not the target cell is in the Surface by using
                the bounding sphere radius that has had cell data appended to it via the converter.

  Arguments:    pCell:      Pointer to the cell information of the decision target
                pMtx:       The coordinate conversion matrix applied to the cell of the decision target
                pViewRect:  Display rectangle of the output destination surface

  Returns:      TRUE if the target cell displays, or FALSE if otherwise.
 *---------------------------------------------------------------------------*/
 
static BOOL CallBackCulling( 
                const NNSG2dCellData*   pCell,
                const MtxFx32*          pMtx,
                const NNSG2dViewRect*   pViewRect)
{
    // Get the bounding sphere radius calculated by the converter
    const fx32  R = NNS_G2dGetCellBoundingSphereR( pCell ) * FX32_ONE;

    //  Look for the cell display position origin
    const fx32  px = pMtx->_20 - pViewRect->posTopLeft.x;
    const fx32  py = pMtx->_21 - pViewRect->posTopLeft.y;
    
    // Look for the rectangle internal to the cell bounding sphere
    const fx32  minY = py - R;
    const fx32  maxY = py + R;
    
    const fx32  minX = px - R;
    const fx32  maxX = px + R;
    
    if( (maxY > 0) && (minY < pViewRect->sizeView.y) )
    {
        if( (maxX > 0) && (minX < pViewRect->sizeView.x) )
        {
            return TRUE;
        }
    }
    return FALSE;
}

/*---------------------------------------------------------------------------*
  Name:         CallBackCulling2

  Description:  This is a function called to determine whether the cell is displayed
                inside the NNS_G2Draw* function.                
                Determines whether the target cell is in the Surface by using the cell
                bounding rectangle that has had cell data appended to it via the converter.
                In order to emphasize speed, the cell affine conversions are not considered.
                
                With this sample, the main screen and the sub screen are being used for titles
                that move up and down.

  Arguments:    pCell:      Pointer to the cell information of the decision target
                pMtx:       The coordinate conversion matrix applied to the cell of the decision target
                pViewRect:  Display rectangle of the output destination surface

  Returns:      TRUE if the target cell displays, or FALSE if otherwise.
 *---------------------------------------------------------------------------*/
 
static BOOL CallBackCulling2(
                const NNSG2dCellData*   pCell,
                const MtxFx32*          pMtx,
                const NNSG2dViewRect*   pViewRect)
{
    // Get the bounding rectangle calculated by the converter
	const NNSG2dCellBoundingRectS16* pRect = NNS_G2dGetCellBoundingRect( pCell );

    // Look for the cell display position origin
    fx32  px = pMtx->_20 - pViewRect->posTopLeft.x;
    fx32  py = pMtx->_21 - pViewRect->posTopLeft.y;

    // Add the cell display position origin
    fx32 min_x  = (pRect->minX << FX32_SHIFT) + px;
    fx32 max_x  = (pRect->maxX << FX32_SHIFT) + px;
    fx32 min_y  = (pRect->minY << FX32_SHIFT) + py;
    fx32 max_y  = (pRect->maxY << FX32_SHIFT) + py;
    
    // Surface inside and outside determination
    if( (max_y > 0) && (min_y < pViewRect->sizeView.y) )
    {
        if( (max_x > 0) && (min_x < pViewRect->sizeView.x) )
        {
            return TRUE;
        }
    }
    return FALSE;
}

/*---------------------------------------------------------------------------*
  Name:         InitRenderer3D

  Description:  Initializes the Renderer and Surface.

  Arguments:    pRenderer:         The pointer to the Renderer to initialize
                pMainSurface3D:    The pointer to the main surface to initialize

  Returns:      None
 *---------------------------------------------------------------------------*/
  
static void InitRenderer3D(
                NNSG2dRendererInstance* pRenderer,
                NNSG2dRenderSurface*    pMainSurface3D)
{
    SDK_NULL_ASSERT( pRenderer );
    SDK_NULL_ASSERT( pMainSurface3D );

    NNS_G2dInitRenderSurface(pMainSurface3D);
        
    // Initiaizes the main screen.
    {
        NNSG2dViewRect* pRect = &(pMainSurface3D->viewRect);

        //  display rectangle	
        pRect->posTopLeft.x = FX32_ONE * 0;
        pRect->posTopLeft.y = FX32_ONE * 0;
        pRect->sizeView.x = FX32_ONE * SCREEN_WIDTH;
        pRect->sizeView.y = FX32_ONE * SCREEN_HEIGHT;
    
        // Callback function 
        // Registration function configuration is not necessary for a surface exclusive to 3D graphics engine rendering
        pMainSurface3D->pFuncOamRegister       = NULL;
        pMainSurface3D->pFuncOamAffineRegister = NULL;
        pMainSurface3D->pFuncVisibilityCulling = CallBackCulling;
        
        //  display destination 
        pMainSurface3D->type = NNS_G2D_SURFACETYPE_MAIN3D;      
    }
  
    // Renderer initialization 
    NNS_G2dInitRenderer( pRenderer );

    // Surface registration 
    NNS_G2dAddRendererTargetSurface( pRenderer, pMainSurface3D );
}

/*---------------------------------------------------------------------------*
  Name:         InitRenderer2D

  Description:  Initializes the Renderer and Surface.

  Arguments:    pRenderer:         The pointer to the Renderer to initialize
                pMainSurface2D:    The pointer to the main surface to initialize
                pSubSurface2D:     The pointer to the sub surface to initialize

  Returns:      None
 *---------------------------------------------------------------------------*/

static void InitRenderer2D(
                NNSG2dRendererInstance* pRenderer2D,
                NNSG2dRenderSurface*    pMainSurface2D,
                NNSG2dRenderSurface*    pSubSurface2D)
{
    SDK_NULL_ASSERT( pRenderer2D );
    SDK_NULL_ASSERT( pMainSurface2D );
    SDK_NULL_ASSERT( pSubSurface2D );

    NNS_G2dInitRenderSurface(pMainSurface2D);
    NNS_G2dInitRenderSurface(pSubSurface2D);

    // Initializes the main screen.
    {
        NNSG2dViewRect* pRect = &(pMainSurface2D->viewRect);

        //  display rectangle
        pRect->posTopLeft = SURFACE_MAIN_INIT_POS;
        pRect->sizeView.x = FX32_ONE * SCREEN_WIDTH;
        pRect->sizeView.y = FX32_ONE * SCREEN_HEIGHT;
    
        // Callback function
        pMainSurface2D->pFuncOamRegister       = CallBackAddOamMain;
        pMainSurface2D->pFuncOamAffineRegister = CallBackAddAffineMain;
        pMainSurface2D->pFuncVisibilityCulling = CallBackCulling2;

        // Display destination
        pMainSurface2D->type = NNS_G2D_SURFACETYPE_MAIN2D;      
    }

    //  Initializes the sub screen.
    {
        NNSG2dViewRect* pRect = &(pSubSurface2D->viewRect);

        //  display rectangle
        pRect->posTopLeft = SURFACE_SUB_INIT_POS;
        pRect->sizeView.x = FX32_ONE * SCREEN_WIDTH;
        pRect->sizeView.y = FX32_ONE * SCREEN_HEIGHT;
    
        // Callback function 
        pSubSurface2D->pFuncOamRegister       = CallBackAddOamSub;
        pSubSurface2D->pFuncOamAffineRegister = CallBackAddAffineSub;
        pSubSurface2D->pFuncVisibilityCulling = CallBackCulling2;

        //  display destination
        pSubSurface2D->type = NNS_G2D_SURFACETYPE_SUB2D;        
    }
  
    // Renderer initialization 
    NNS_G2dInitRenderer( pRenderer2D );

    // Surface registration 
    NNS_G2dAddRendererTargetSurface( pRenderer2D, pMainSurface2D );
    NNS_G2dAddRendererTargetSurface( pRenderer2D, pSubSurface2D );
}

/*---------------------------------------------------------------------------*
  Name:         CallBackAddOam*

  Description:  This is a function called to add the OAM in the NNS_G2dDraw* function.

  Arguments:    pOam:           the pointer to the OAM to add.
                affineIndex:    the affine index this OAM uses.
                                It can have a value greater than 32.
                                If the affine is not used,
                                designate NNS_G2D_OAM_AFFINE_IDX_NONE.
                bDoubleAffine:  whether or not it s double affine mode

  Returns:      Returns TRUE if successful, or FALSE if otherwise.
 *---------------------------------------------------------------------------*/

static BOOL CallBackAddOamMain( const GXOamAttr* pOam, u16 affineIndex, BOOL /*bDoubleAffine*/ )
{
    SDK_NULL_ASSERT( pOam );

    return NNS_G2dEntryOamManagerOamWithAffineIdx( &sMainOamMan, pOam, affineIndex );
}

static BOOL CallBackAddOamSub( const GXOamAttr* pOam, u16 affineIndex, BOOL /*bDoubleAffine*/ )
{
    SDK_NULL_ASSERT( pOam );
        
    return NNS_G2dEntryOamManagerOamWithAffineIdx( &sSubOamMan, pOam, affineIndex ); 
}

/*---------------------------------------------------------------------------*
  Name:         CallBackAddAffine*

  Description:  This is a function called to add an affine parameter inside the NNS_G2dDraw* function.
                It passes the affine parameters to the OAM manager.

  Arguments:    mtx:        the pointer to the affine variable matrix to be added 

  Returns:      Returns the AffineIndex for the added affine paremeter
                to reference.
 *---------------------------------------------------------------------------*/

static u16 CallBackAddAffineMain( const MtxFx22* mtx )
{
    SDK_NULL_ASSERT( mtx );
    return NNS_G2dEntryOamManagerAffine( &sMainOamMan, mtx );
}

static u16 CallBackAddAffineSub( const MtxFx22* mtx )
{
    SDK_NULL_ASSERT( mtx );
    return NNS_G2dEntryOamManagerAffine( &sSubOamMan, mtx ); 
}

/*---------------------------------------------------------------------------*
  Name:         CallbackRegistTransferTask

  Description:  This is a function called to register the transmission task of the
                character data inside the NNS_G2dDraw function.

  Arguments:    type:        transmission destination VRAM  region type
                dstAddr:     transmission destination address  offset 
                pSrc:        pointer to the transmission source data
                szByte:      the data size that should be transmitted

  Returns:      Returns TRUE if successful, or FALSE if otherwise.
 *---------------------------------------------------------------------------*/

static BOOL CallbackRegistTransferTask(
                NNS_GFD_DST_TYPE type,
                u32 dstAddr,
                void* pSrc,
                u32 szByte )
{
    // Passes as-is to the VRAM transmission manager
    return NNS_GfdRegisterNewVramTransferTask(
                type,
                dstAddr,
                pSrc,
                szByte );
}

/*---------------------------------------------------------------------------*
  Name:         InitG2D

  Description:  Performs G2D-related initialization

  Arguments:    None

  Returns:      None
 *---------------------------------------------------------------------------*/

void InitG2D(void)
{
    // OamManager initialization 
    NNS_G2dInitOamManagerModule();
    InitOamManager( &sMainOamMan, NNS_G2D_OAMTYPE_MAIN );
    InitOamManager( &sSubOamMan, NNS_G2D_OAMTYPE_SUB );

    // Renderer initialization
    InitRenderer3D( &render3D, &mainSurface3D);
    InitRenderer2D( &render2D, &mainSurface2D , &subSurface2D);

    NNS_G2dSetupSoftwareSpriteCamera();

    // Configures the attributes that become valid when rendering.
    // The performance will increase if unnecessary attributes are configured as invalid.

    NNS_G2dSetSpriteAttrEnable( NNS_G2D_SPRITEATTR_COLOR | 
                                NNS_G2D_SPRITEATTR_ALPHA | 
                                NNS_G2D_SPRITEATTR_UV    | 
                                NNS_G2D_SPRITEATTR_ROTXY | 
                                NNS_G2D_SPRITEATTR_TEXTUE|
                                NNS_G2D_SPRITEATTR_ROTO  |
                                NNS_G2D_SPRITEATTR_FLIP );  
    
    // Perform initialization for VRAM transmission
    {
        // Initialize the VRAM transmission manager.
        NNS_GfdInitVramTransferManager(
            sTransferTaskArray,                 // pointer to the VRAM transmission task array
            NUM_OF_VRAMTASK                     // VRAM transmission task array length
        );
            
        //  Initialize the cell VRAM transmission state object manager
        NNS_G2dInitCellTransferStateManager(
            sVramStateArray,                    //  cell VRAM transmission state object buffer 
            NUM_OF_VRAMSTATE,                   //  cell VRAM transmission state object buffer length
            CallbackRegistTransferTask          // VRAM transmission registration function pointer 
        );
    }    
}

