/*---------------------------------------------------------------------------*
  Project:  G2D Program Sample
  File:     SceneStage00.c

  Copyright 2004 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: $
  $NoKeywords: $
 *---------------------------------------------------------------------------*/

#include <nitro.h>
#include <nnsys/fnd.h>
#include <nnsys/g2d.h>
#include "utility.h"
#include "task_priority.h"
#include "vram_map.h"
#include "read_key.h"
#include "yoshi.h"
#include "enemy.h"
#include "effect.h"
#include "background.h"
#include "renderer.h"
#include "scene.h"
#include "fade.h"
#include "schedule.h"

// Declaration of functions
extern void CheckCollision(void);
static void LoadData(void);

static void 
LoadNitroCharacterData(
    ObjData* objData, 
    const NitroCharacterData* nitroCharData,
    u32 baseAddr
);

static void 
LoadNitroCharacterDataVramTransfer(
    ObjData* objData, 
    const NitroCharacterData* nitroCharData,
    u32 baseAddr
);

static void SceneStage00_Free(void);
static void SceneStage00_DrawInit(void);

/*---------------------------------------------------------------------------*
  Name:         SceneStage00_Init

  Description:  Performs the initial stage 00 configuration.
                        
  Arguments:    None
  
  Returns:      None
 *---------------------------------------------------------------------------*/

void SceneStage00_Init(void)
{
    Stage00_ClearEventFlags();          // Clear event flag
    SceneStage00_DrawInit();            // Render screen setup   
    CameraSetup();                      // Camera setup
    MaterialSetup();                    // Material setup
    LoadData();                         // Load data
    
    // Stage 00 initial task configuration
    {
        Task* pSelfTask = (Task *)TaskCtrl_GetCurrentTask();
        YoshiTask*  yoshi;
        EffectTask* effect;
        // Key input determination task
        (void)TaskCtrl_SetTask(ReadKey, TASK_PRIORITY_READKEY);
        // Process for before and after Render
        (void)TaskCtrl_SetTask(BeginRendering2D, TASK_PRIORITY_BEGIN_RENDER2D);
        (void)TaskCtrl_SetTask(EndRendering2D,   TASK_PRIORITY_END_RENER2D);        
        (void)TaskCtrl_SetTask(BeginRendering3D, TASK_PRIORITY_BEGIN_RENDER3D);
        (void)TaskCtrl_SetTask(EndRendering3D,   TASK_PRIORITY_END_RENER3D);        
        // Yoshi task 
        yoshi = (YoshiTask *)TaskCtrl_SetTask(YoshiBirth, TASK_PRIORITY_YOSHI);
        yoshi->pos.x = FX32_SCREEN_CENTER_X;      
        yoshi->pos.y = FX32_ONE * (192-48);
        yoshi->direction = RIGHT;
        // Collision determination task
        (void)TaskCtrl_SetTask(CheckCollision, TASK_PRIORITY_COLLISION);
        // Scheduler task
        (void)TaskCtrl_SetTask(Stage00_Schedule, TASK_PRIORITY_SCHEDULE);    
        // Background task
        (void)TaskCtrl_SetTask(BackGround01_Init, TASK_PRIORITY_BG);
        (void)TaskCtrl_SetTask(BackGround02_Init, TASK_PRIORITY_BG);
        (void)TaskCtrl_SetTask(BackGround03_Init, TASK_PRIORITY_BG);
        // Pipe task
        effect = (EffectTask *)TaskCtrl_SetTask(Pipe, TASK_PRIORITY_EFFECT_NOTRENDER);
        effect->pos.x = FX32_ONE * 1142;
        effect->pos.y = FX32_ONE * 113;
        effect->size.x = FX32_ONE * 32;
        effect->size.y = FX32_ONE * 32;
        effect->rotate = 0;
        // Mini title task
        (void)TaskCtrl_SetTask(TitleMini_Init, TASK_PRIORITY_EFFECT_NOTRENDER);  // Add a title task
        // Fade in task
        (void)TaskCtrl_SetTask(FadeIn, TASK_PRIORITY_FADEIN);       
        
        pSelfTask->common.releaseFunc = SceneStage00_Free;       // Set the deallocation function
        pSelfTask->common.func = SceneStage00;                   // Change task
    }
}

/*---------------------------------------------------------------------------*
  Name:         LoadData

  Description:  Loads the character data, cell data, animation data, and palette data
                needed in Stage00, and if necessary loads them into VRAM.
                Also performs animation initialization as needed.
                As it now stands, the process is falling off because the data is being loaded all at once.
                Try to come up with some means of splitting up the data before loading to prevent this processing problem.
                        
  Arguments:    None
  
  Returns:      None
 *---------------------------------------------------------------------------*/

static void LoadData(void)
{
    // Yoshi
    UT_SetUpNitroCharacterData3D(
        &objDataYoshi, 
        &fileYoshi, 
        VRAM_MAP_TEX_YOSHI, 
        VRAM_MAP_TEXPLTT_YOSHI);
    // Heiho
    UT_SetUpNitroCharacterData3D(
        &objDataHeihoRed, 
        &fileHeihoRed, 
        VRAM_MAP_TEX_HEIHO_RED, 
        VRAM_MAP_TEXPLTT_HEIHO_RED);
    // Smoke
    UT_SetUpNitroCharacterData3D(
        &objDataSmoke, 
        &fileSmoke, 
        VRAM_MAP_TEX_SMOKE, 
        VRAM_MAP_TEXPLTT_SMOKE);
    // Pipe
    UT_SetUpNitroCharacterData3D(
        &objDataPipe,  
        &filePipe, 
        VRAM_MAP_TEX_PIPE, 
        VRAM_MAP_TEXPLTT_PIPE);
    // Star (damage effect)
    UT_SetUpNitroCharacterData3D(
        &objDataStar,  
        &fileStar, 
        VRAM_MAP_TEX_STAR, 
        VRAM_MAP_TEXPLTT_STAR);
    // Killer
    UT_SetUpNitroCharacterData3D(
        &objDataKiller,  
        &fileKiller, 
        VRAM_MAP_TEX_KILLER, 
        VRAM_MAP_TEXPLTT_KILLER);
    // Mini title
    UT_SetUpNitroCharacterData3D(
        &objDataTitleMini,  
        &fileTitleMini, 
        VRAM_MAP_TEX_TITLE_MINI, 
        VRAM_MAP_TEXPLTT_TITLE_MINI);       
}

/*---------------------------------------------------------------------------*
  Name:         SceneStage00_Free

  Description:  Deallocates the region alocated with SceneStage00_Init.

  Arguments:    None

  Returns:      None
 *---------------------------------------------------------------------------*/

static void SceneStage00_Free(void)
{
    UT_Free( objDataYoshi.pFileChar ); 
    UT_Free( objDataYoshi.pFileCell ); 
    UT_Free( objDataYoshi.pFileAnim ); 
    UT_Free( objDataHeihoRed.pFileCell ); 
    UT_Free( objDataHeihoRed.pFileAnim );   
    UT_Free( objDataStar.pFileCell );
    UT_Free( objDataStar.pFileAnim );
}

/*---------------------------------------------------------------------------*
  Name:         SceneStage00_DrawInit

  Description:  Performs render-related configurations.

  Arguments:    None

  Returns:      None
 *---------------------------------------------------------------------------*/

static void SceneStage00_DrawInit(void)
{
    GX_SetVisiblePlane(   GX_PLANEMASK_BG0 |
                          GX_PLANEMASK_BG1 |
                          GX_PLANEMASK_BG2 |
                          GX_PLANEMASK_BG3 |
                          GX_PLANEMASK_OBJ );
}


