/*---------------------------------------------------------------------------*
  Project:  G2D Program Sample
  File:     scene_title_init.c

  Copyright 2004 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: $
  $NoKeywords: $
 *---------------------------------------------------------------------------*/

#include <nitro.h>
#include <nnsys/g2d.h>
#include "utility.h"
#include "task_priority.h"
#include "vram_map.h"
#include "read_key.h"
#include "effect.h"
#include "scene.h"
#include "fade.h"
#include "renderer.h"

// Declaration of functions
static void SceneTitle_Free(void);
static inline void SceneTitle_DrawInit(void);
static void LoadData(void);

// Declaration of internal variables
static BOOL firstTime = TRUE;   // This is set to FALSE the first time this task is executed.

/*---------------------------------------------------------------------------*
  Name:         SceneTitle_Init

  Description:  Performs the initial configuration on the title sequence.
                        
  Arguments:    None
  
  Returns:      None
 *---------------------------------------------------------------------------*/

void SceneTitle_Init(void)
{
    SceneTask* pSelfTask = (SceneTask *)TaskCtrl_GetCurrentTask();

    // Execute only the first time
    if (firstTime == TRUE)
    {
        Task* task;
        LoadData();                                               // Load data
        task = TaskCtrl_SetTask(Title_Init, TASK_PRIORITY_TITLE); // Title display task
        pSelfTask->relateTask = task;                             // Set the related tasks to the title display task
        firstTime = FALSE;  
    }
    
    // Process before and after Render
    (void)TaskCtrl_SetTask(BeginRendering2D, TASK_PRIORITY_BEGIN_RENDER2D);
    (void)TaskCtrl_SetTask(EndRendering2D,   TASK_PRIORITY_END_RENER2D);        

    SceneTitle_DrawInit();                                      // Render screen configuration   
    CameraSetup();                                              // Camera configuration
    MaterialSetup();                                            // Material configuration 

    *(GXRgb*)(HW_BG_PLTT)    = COLOR_BLACK;                     // Set the backdrop color to black (main screen)
    *(GXRgb*)(HW_DB_BG_PLTT) = COLOR_BLACK;                     // Set the backdrop color to black (sub screen)

    (void)TaskCtrl_SetTask(ReadKey, TASK_PRIORITY_READKEY);     // key input determination task
    (void)TaskCtrl_SetTask(FadeIn, TASK_PRIORITY_FADEIN);       // Fade in task

    pSelfTask->common.releaseFunc = SceneTitle_Free;            // Deallocation function set
    pSelfTask->common.func = SceneTitle;                        // Change task
}

/*---------------------------------------------------------------------------*
  Name:         LoadData

  Description:  Load the character data, cell data, animation data, and palette data
                needed for the title sequence, and load into VRAM if necessary.
                Also, perform the initialization for the animation
                if necessary.
                        
  Arguments:    None
  
  Returns:      None
 *---------------------------------------------------------------------------*/

static void LoadData(void)
{
    UT_SetUpNitroCharacterData2D(
        &objDataTitle,
        &fileTitle,
        VRAM_MAP_TEX_TITLE,
        VRAM_MAP_TEXPLTT_TITLE
    );
}

/*---------------------------------------------------------------------------*
  Name:         SceneTitle_Free

  Description:  Deallocates the region allocated with SceneTitle_Init.

  Arguments:    None

  Returns:      None
 *---------------------------------------------------------------------------*/

static void SceneTitle_Free(void)
{
    
}

/*---------------------------------------------------------------------------*
  Name:         SceneTitle_DrawInit

  Description:  Performs render-related configuration.

  Arguments:    None

  Returns:      None
 *---------------------------------------------------------------------------*/

static inline void SceneTitle_DrawInit(void)
{
    GX_SetVisiblePlane( GX_PLANEMASK_OBJ );
}

