/*---------------------------------------------------------------------------*
  Project:  G2D Program Sample
  File:     task_control.c

  Copyright 2004 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: $
  $NoKeywords: $
 *---------------------------------------------------------------------------*/
 
#include "task_control.h"

// Internal variable definition
static Task  sTask[TASK_CTRL_AREA_NUM];     /* Task array */
static Task* TaskStack[TASK_CTRL_AREA_NUM]; /* Task stack */
static Task** pTaskStack = &TaskStack[0];   /* Task stack pointer */
static Task*  pCurrentTask;                 /* Task that is currently being executed */

// Internal function declaration
static void TaskCtrl_ReleaseTasks(void);

/*---------------------------------------------------------------------------*
  Name:         TaskCtrl_Init

  Description:  Initializes the task control
                        
  Arguments:    None
                
  Returns:      None
 *---------------------------------------------------------------------------*/

void TaskCtrl_Init(void)
{
    u32 i;
    
    for (i=0;i<TASK_CTRL_AREA_NUM;i++)
    {
        TaskStack[i] = &sTask[i];
    }
    
    // Beginning task registration
    TaskStack[TASK_NUMBER_HEAD] = NULL;                                 // Empty task clear
    pTaskStack++;                                                       // Task stack pointer increment
    sTask[TASK_NUMBER_HEAD].common.func     = OperateNothing;           // Execution task
    sTask[TASK_NUMBER_HEAD].common.priority = 0;                        // Highest priority
    sTask[TASK_NUMBER_HEAD].common.prev     = 0;                        // The precedent task does not exist 
    sTask[TASK_NUMBER_HEAD].common.next     = &sTask[TASK_NUMBER_TAIL]; // The next task is the subsequent task
    
    // Subsequent task registration
    TaskStack[TASK_NUMBER_TAIL] = NULL;                                 // Empty task clear
    pTaskStack++;                                                       // Ring buffer pointer increment 
    sTask[TASK_NUMBER_TAIL].common.func     = OperateNothing;           // Execution task
    sTask[TASK_NUMBER_TAIL].common.priority = 0xffff;                   // Lowest priority
    sTask[TASK_NUMBER_TAIL].common.prev     = &sTask[TASK_NUMBER_HEAD]; // The precedent task is the beginning task
    sTask[TASK_NUMBER_TAIL].common.next     = NULL;                     // The next task does not exist
    
    // Task that is currently being executed
    pCurrentTask = &sTask[TASK_NUMBER_HEAD];
}

/*---------------------------------------------------------------------------*
  Name:         TaskCtrl_SetTask

  Description:  Registers tasks
                        
  Arguments:    func:      Function to execute
                priority:  Priority
                
  Returns:      Pointer to the registered task
 *---------------------------------------------------------------------------*/

Task* TaskCtrl_SetTask(Funcp func, u16 priority)
{
    Task* pNext = TaskCtrl_GetHeadTask();
    
    if (pTaskStack == &TaskStack[TASK_CTRL_AREA_NUM - 1])
    {
        OS_Panic("Fail TaskCtrl_SetTask");                  // Fails because there is no opening in task
    }

    (*pTaskStack)->common.func     = func;                  // Execution function set
    (*pTaskStack)->common.priority = priority;              // Priority set

    while (1) {
        if (pNext->common.priority > priority)              // Search for higher priority
        {
            (*pTaskStack)->common.next = pNext;             // Insert the task in doubly linked list
            (*pTaskStack)->common.prev = pNext->common.prev;// 
            pNext->common.prev->common.next   = *pTaskStack;// 
            pNext->common.prev         = *pTaskStack;       //          
            break;
        }
        pNext = pNext->common.next;                         // To the next task
    }
    
    *pTaskStack = NULL;                                     // Used stack area clear
    pTaskStack++;                                           // Stack pointer increment

    return pNext->common.prev;
}

/*---------------------------------------------------------------------------*
  Name:         TaskCtrl_ExecuteAllTasks

  Description:  Executes all tasks.
                Then calls TaskCtrl_ReleaseTasks.
                        
  Arguments:    None
                
  Returns:      None
 *---------------------------------------------------------------------------*/
 
void TaskCtrl_ExecuteAllTasks(void)
{
    Task  *current = TaskCtrl_GetHeadTask();

    while(1)
    {
        if (current->common.next == NULL)
        {
            break;                          // Escape if subsequent task
        }
        current = current->common.next;     // To the next task
        pCurrentTask = current;             // After changing the current task 
        current->common.func();             // Execute
    }
    pCurrentTask = TaskCtrl_GetHeadTask();  // Return the current task to the beginning
    TaskCtrl_ReleaseTasks();                // Release the task that has the release flag enabled
}

/*---------------------------------------------------------------------------*
  Name:         TaskCtrl_ReleaseTasks

  Description:  Releases the task that has the release request flag enabled.
                        
  Arguments:    None
                
  Returns:      None
 *---------------------------------------------------------------------------*/

static void TaskCtrl_ReleaseTasks(void)
{   
    Task  *pRelease = TaskCtrl_GetHeadTask();
    
    while (1)
    {
        Task  *pNext = pRelease->common.next;
        
        if (pRelease->common.reqReleaseFlag == TRUE)
        {
            pRelease->common.prev->common.next = pRelease->common.next;  // Reconnecting links
            pRelease->common.next->common.prev = pRelease->common.prev;

            if (pRelease->common.releaseFunc != NULL)
            {
                pRelease->common.releaseFunc(pRelease);  // Release memory
            }
            
            MI_CpuClear8( pRelease, sizeof(Task));       // Clear task
    
            pTaskStack--;
            *pTaskStack = pRelease;                      // Return to stack
        }
        
        pRelease = pNext;
        
        if (pRelease->common.next == NULL)              // End if subsequent task
        {
            break;
        }
    }
}

/*---------------------------------------------------------------------------*
  Name:         TaskCtrl_SetReleaseFlagAll

  Description:  Enables the release request flag for all tasks excluding HEAD and TAIL.
                        
  Arguments:    None
                
  Returns:      None
 *---------------------------------------------------------------------------*/

void TaskCtrl_SetReleaseFlagAll(void)
{
    Task  *pRelease = TaskCtrl_GetHeadTask()->common.next;
    
    while (1)
    {
        if (pRelease->common.next == NULL)     // End if TAIL
        {
            break;
        }   
        pRelease->common.reqReleaseFlag = TRUE;
        pRelease = pRelease->common.next;
    }
}

/*---------------------------------------------------------------------------*
  Name:         TaskCtrl_SetReleaseFlagRange

  Description:  Enables the release request flag of the task by specifying the range with priority.
  
                        
  Arguments:    u16 minimum   The minimum value of the range
                u16 maximum   The maximum value of the range
                
  Returns:      None
 *---------------------------------------------------------------------------*/

void TaskCtrl_SetReleaseFlagRange(u16 minimum, u16 maximum)
{   
    Task  *pRelease = TaskCtrl_GetHeadTask()->common.next;
    
    // Search for the beginning task in the range
    while (1)
    {
        if (pRelease->common.priority >= minimum)
        {
            break;      // The Beginning task found
        }
        pRelease = pRelease->common.next;       
    }
    
    // Releases the task in the range
    while (1)
    {
        if (pRelease->common.priority > maximum || pRelease->common.next == NULL)
        {
            break;     // End if outside the range or the TAIL task
        }   
        pRelease->common.reqReleaseFlag = TRUE;
        pRelease = pRelease->common.next;
    }
}

/*---------------------------------------------------------------------------*
  Name:         TaskCtrl_SetReleaseFlagByPriority

  Description:  Enables the release request flag of the corresponding task by specifying priority.
  
                        
  Arguments:    u16 priority
                
  Returns:      None
 *---------------------------------------------------------------------------*/

void TaskCtrl_SetReleaseFlagByPriority(u16 priority)
{
    TaskCtrl_SetReleaseFlagRange(priority, priority);
}

/*---------------------------------------------------------------------------*
  Name:         TaskCtrl_SetReleaseFlagAllWithMask

  Description:  Enables the release request flag for all tasks other excluding HEAD and TAIL.
                However, the mask can be specified with the priority value.
                        
  Arguments:    u16 mask    Priority (array) of that task which the release flag is not enabled.
                u16 num     Number of elements in array
                
  Returns:      None
 *---------------------------------------------------------------------------*/

void TaskCtrl_SetReleaseFlagAllWithMask(u16 mask[], u16 num)
{   
    Task  *pRelease = TaskCtrl_GetHeadTask()->common.next;
    BOOL  masked;
    u32   i;
    
    while (1)
    {
        if (pRelease->common.next == NULL)     // End if TAIL
        {
            break;
        }
        
        // Determines if it is masked
        masked = FALSE;
        for (i=0;i<num;i++)
        {
            if (pRelease->common.priority == mask[i])
            {
                masked = TRUE;
                break;
            }
        }
        
        // If not masked, enabled the release flag
        if (masked == FALSE)
        {
            pRelease->common.reqReleaseFlag = TRUE;
        }
        pRelease = pRelease->common.next;
    }
}

/*---------------------------------------------------------------------------*
  Name:         TaskCtrl_GetCurrentTask

  Description:  Returns a pointer to the task that is currently processing.
                        
  Arguments:    None
                
  Returns:      Pointer to the task that is currently processing.
 *---------------------------------------------------------------------------*/
 
Task* TaskCtrl_GetCurrentTask(void)
{
    return pCurrentTask;
}

/*---------------------------------------------------------------------------*
  Name:         TaskCtrl_GetHeadTask

  Description:  Returns a pointer to the beginning task.
                        
  Arguments:    None
                
  Returns:      Pointer to the beginning task
 *---------------------------------------------------------------------------*/

Task* TaskCtrl_GetHeadTask(void)
{
    return &sTask[TASK_NUMBER_HEAD];
}

/*---------------------------------------------------------------------------*
  Name:         OperateNothing

  Description:  Returns without doing anything
                        
  Arguments:    None
                
  Returns:      None
 *---------------------------------------------------------------------------*/

void OperateNothing(void){}

