/*---------------------------------------------------------------------------*
  Project:  G2D Program Sample
  File:     utility.c

  Copyright 2004 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: $
  $NoKeywords: $
 *---------------------------------------------------------------------------*/

#include <nitro.h>
#include <nnsys/g2d.h>
#include "utility.h"
#include "yoshi.h"

/* Cell animation entity */
static NNSG2dCellAnimation   CellAnimArray[CELL_ANIM_ARRAY_SIZE];
static NNSG2dCellAnimation*  CellAnimStack[CELL_ANIM_ARRAY_SIZE];
static NNSG2dCellAnimation** pCellAnimStack = &CellAnimStack[0];

/*---------------------------------------------------------------------------*
  Name:         UT_InitCellAnimStack

  Description:  Initializes the stack used for obtaining a cell animation.

  Arguments:    None
  
  Returns:      None
 *---------------------------------------------------------------------------*/

void UT_InitCellAnimStack(void)
{
    u32 i;
    
    for (i=0;i<CELL_ANIM_ARRAY_SIZE; i++)
    {
        CellAnimStack[i] = &CellAnimArray[i];
    }
}

/*---------------------------------------------------------------------------*
  Name:         UT_GetCellAnimation

  Description:  Obtains a cell animation from stack.

  Arguments:    None
  
  Returns:      None
 *---------------------------------------------------------------------------*/

NNSG2dCellAnimation* UT_GetCellAnimation(void)
{   
    NNSG2dCellAnimation* ret;
    
    if (pCellAnimStack == &CellAnimStack[CELL_ANIM_ARRAY_SIZE-1])
    {
        OS_Panic("Fail TaskCtrl_SetTask");
    }
    
    ret = *pCellAnimStack;
    pCellAnimStack++;
    
    return ret;
}

/*---------------------------------------------------------------------------*
  Name:         UT_ReleaseCellAnimation

  Description:  Returns a cell animation to stack.

  Arguments:    None
  
  Returns:      None
 *---------------------------------------------------------------------------*/

void UT_ReleaseCellAnimation(NNSG2dCellAnimation* anim)
{
    pCellAnimStack--;
    *pCellAnimStack = anim;
}


/*---------------------------------------------------------------------------*
  Name:         UT_LoadFileToMainMem

  Description:  Allocates buffer and loads file.

  Arguments:    pFname:        Name of the file to load

  Returns:      If loading was successful, it returns the pointer to the buffer 
                that stores the file content. This buffer needs to be released  
                with UT_Free when no longer needed.
                If loading fails, it returns NULL.
 *---------------------------------------------------------------------------*/
 
static void* UT_LoadFileToMainMem( const char* pFname )
{
    FSFile  file;
    void*   pFile = NULL;
    BOOL    bSuccess;

    SDK_NULL_ASSERT( pFname );

    FS_InitFile( &file );
    bSuccess = FS_OpenFile( &file, pFname );
    
    if( bSuccess )
    {
        const u32 szFile    = FS_GetLength( &file );

        pFile = NNS_FndAllocFromExpHeap(gHeapHandle, szFile );
        SDK_NULL_ASSERT( pFile );

        if( szFile != FS_ReadFile( &file, pFile, szFile ) )
        {
            NNS_FndFreeToExpHeap(gHeapHandle, pFile );
            pFile = NULL;
        }

        bSuccess = FS_CloseFile( &file );
        SDK_ASSERT( bSuccess );
    }else{
        OS_Warning(" Can't find the file : %s ", pFname );
    }

    return pFile;
}

/*---------------------------------------------------------------------------*
  Name:         UT_LoadNCER

  Description:  Internally allocates buffer and loads cell data bank from NCER file.
                After the function succeeds, the buffer must be released with UT_Free (return value) 
                when the cell data bank is no longer needed.


  Arguments:    ppCellBank: Pointer that receives a pointer to the buffer that stores cell data bank

                pFname:     Name of the NCER file that stores cell data bank

  Returns:      If loading was successful, it returns a pointer to the internally allocated buffer 
                that stores the file data.
                If loading fails, it returns NULL.
 *---------------------------------------------------------------------------*/
 
void* UT_LoadNCER( NNSG2dCellDataBank** ppCellBank, const char* pFname )
{
    void*   pFile = NULL;

    SDK_NULL_ASSERT( ppCellBank );
    SDK_NULL_ASSERT( pFname );

    pFile = UT_LoadFileToMainMem( pFname );
    if( pFile != NULL )
    {
        if( NNS_G2dGetUnpackedCellBank( pFile, ppCellBank ) )
        {
            NNS_G2dPrintCellBank( *ppCellBank );
            return pFile;
        }
        UT_Free(pFile);
    }

    return NULL;
}

/*---------------------------------------------------------------------------*
  Name:         UT_LoadNANR

  Description:  Internally allocates buffer and loads animation data bank from NANR file.
                After the function succeeds, the buffer must be released with 
                UT_Free (return value) when the animation data bank is no longer needed.


  Arguments:    ppAnimBank: Pointer that receives a pointer to the buffer that stores animation data bank

                pFname:     Name of the NANR file that stores animation data bank


  Returns:      If loading was successful, it returns a pointer to the internally allocated buffer 
                that stores the file data.
                If loading fails, it returns NULL.
 *---------------------------------------------------------------------------*/
 
void* UT_LoadNANR( NNSG2dAnimBankData** ppAnimBank, const char* pFname )
{
    void*   pFile = NULL;

    SDK_NULL_ASSERT( ppAnimBank );
    SDK_NULL_ASSERT( pFname );

    pFile = UT_LoadFileToMainMem( pFname );
    if( pFile != NULL )
    {
        if( NNS_G2dGetUnpackedAnimBank( pFile, ppAnimBank ) )
        {
            NNS_G2dPrintAnimBank( *ppAnimBank );
            return pFile;
        }
    
        UT_Free(pFile);

    }

    return NULL;
}
/*---------------------------------------------------------------------------*
  Name:         UT_LoadNMCR

  Description:  Internally allocates buffer and loads multicell data bank from NMCR file
                After the function succeeds, the buffer must be released with 

                UT_Free (return value) when the multicell data bank is no longer needed.

  Arguments:    ppMCBank:   Pointer that receives a pointer to the buffer that stores multicell data bank

                pFname:     Name of the NMCR file that stores multicell data bank


  Returns:      If loading was successful, it returns a pointer to the internally allocated buffer 
                that stores the file data.
                If loading fails, it returns NULL.
 *---------------------------------------------------------------------------*/
 
void* UT_LoadNMCR( NNSG2dMultiCellDataBank** ppMCBank, const char* pFname )
{
    void*   pFile = NULL;

    SDK_NULL_ASSERT( ppMCBank );
    SDK_NULL_ASSERT( pFname );

    pFile = UT_LoadFileToMainMem( pFname );
    if( pFile != NULL )
    {
        if( NNS_G2dGetUnpackedMultiCellBank( pFile, ppMCBank ) )
        {
            NNS_G2dPrintMultiCellBank ( *ppMCBank );
            return pFile;
        }

        UT_Free(pFile);
    }

    return NULL;
}

/*---------------------------------------------------------------------------*
  Name:         UT_LoadNMAR

  Description:  Internally allocates buffer, and loads multicell animation data bank from NMAR file
                After the function succeeds, the buffer must be released with 

                UT_Free (return value) when the multi animation data bank is no longer needed.

  Arguments:    ppMCABank:  Pointer that receives a pointer to the buffer that stores multicell animation data bank

                pFname:     Name of the NMAR file that stores multicell animation data bank


  Returns:      If loading was successful, it returns a pointer to the internally allocated buffer 
                that stores the file data.
                If loading fails, it returns NULL.
 *---------------------------------------------------------------------------*/
 
void* UT_LoadNMAR( NNSG2dMultiCellAnimBankData** ppMCABank, const char* pFname )
{
    void*   pFile = NULL;

    SDK_NULL_ASSERT( ppMCABank );
    SDK_NULL_ASSERT( pFname );

    pFile = UT_LoadFileToMainMem( pFname );
    if( pFile != NULL )
    {
        if( NNS_G2dGetUnpackedMCAnimBank( pFile, ppMCABank ) )
        {
            NNS_G2dPrintAnimBank( *ppMCABank );
            return pFile;
        }
    
        UT_Free(pFile);
    }

    return NULL;
}

/*---------------------------------------------------------------------------*
  Name:         UT_LoadNCGR

  Description:  Internally allocates buffer, and loads character data in character format from NCGR file
                After the function succeeds, the buffer must be released with 

                UT_Free (return value) when the character data is no longer needed.

  Arguments:    ppCharData: Pointer that receives a pointer to the buffer that stores character data

                pFname:     Name of the NCGR file that stores character data

  Returns:       If loading was successful, it returns a pointer to the internally allocated buffer 
                that stores the file data.
                If loading fails, it returns NULL.
 *---------------------------------------------------------------------------*/
 
void* UT_LoadNCGR( NNSG2dCharacterData** ppCharData, const char* pFname )
{
    void*   pFile = NULL;

    SDK_NULL_ASSERT( ppCharData );
    SDK_NULL_ASSERT( pFname );

    pFile = UT_LoadFileToMainMem( pFname );
    if( pFile != NULL )
    {
        if( NNS_G2dGetUnpackedCharacterData( pFile, ppCharData ) )
        {
            NNS_G2dPrintCharacterData( *ppCharData );
            return pFile;
        }

        UT_Free(pFile);

    }

    return NULL;
}

/*---------------------------------------------------------------------------*
  Name:         UT_LoadNCLR

  Description:  Internally allocates buffer, and loads palette data from NCLR file
                After the function succeeds, the buffer must be released with 

                UT_Free (return value) when the palette data is no longer needed.

  Arguments:    ppPltData: Pointer that receives a pointer to the buffer that stores palette data

                pFname:     Name of the NCLR file that stores palette data

  Returns:      If loading was successful, it returns a pointer to the internally allocated buffer 
                that stores the file data.
                If loading fails, it returns NULL.
 *---------------------------------------------------------------------------*/
 
void* UT_LoadNCLR( NNSG2dPaletteData** ppPltData, const char* pFname )
{
    void*   pFile = NULL;

    SDK_NULL_ASSERT( ppPltData );
    SDK_NULL_ASSERT( pFname );

    pFile = UT_LoadFileToMainMem( pFname );
    if( pFile != NULL )
    {
        if( NNS_G2dGetUnpackedPaletteData( pFile, ppPltData ) )
        {
            NNS_G2dPrintPaletteData( *ppPltData );
            return pFile;
        }
        UT_Free(pFile);
    }

    return NULL;
}

/*---------------------------------------------------------------------------*
  Name:         UT_LoadNENR

  Description:  Internally allocates buffer, and loads entity data from NENR file
                After the function succeeds, the buffer must be released with 

                UT_Free (return value) when the entity data bank is no longer needed.

  Arguments:    ppEntityBank: Pointer that receives a pointer to the buffer that stores entity data bank

                pFname:     Name of the NENR file that stores entity data bank


  Returns:      If loading was successful, it returns a pointer to the internally allocated buffer 
                that stores the file data.
                If loading fails, it returns NULL.
 *---------------------------------------------------------------------------*/
 
void* UT_LoadNENR( NNSG2dEntityDataBank** ppEntityBank, const char* pFname )
{
    void*   pFile = NULL;

    SDK_NULL_ASSERT( ppEntityBank );
    SDK_NULL_ASSERT( pFname );

    pFile = UT_LoadFileToMainMem( pFname );
    if( pFile != NULL )
    {
        if( NNS_G2dGetUnpackedEntityBank( pFile, ppEntityBank ) )
        {
            NNS_G2dPrintEntityDataBank( *ppEntityBank );
            return pFile;
        }
    
        UT_Free(pFile);
    }

    return NULL;
}

/*---------------------------------------------------------------------------*
  Name:         UT_LoadNCBR

  Description:  Internally allocates buffer, and loads character data in bitmap format from NCBR file
                After the function succeeds, the buffer must be released with 

                UT_Free (return value) when the character data is no longer needed.

  Arguments:    ppCharData: Pointer that receives a pointer to the buffer that stores character data

                pFname:     Name of the NCBR file that stores character data

  Returns:      If loading was successful, it returns a pointer to the internally allocated buffer 
                that stores the file data.
                If loading fails, it returns NULL.
 *---------------------------------------------------------------------------*/
 
void* UT_LoadNCBR( NNSG2dCharacterData** ppCharData, const char* pFname )
{
    void*   pFile = NULL;

    SDK_NULL_ASSERT( ppCharData );
    SDK_NULL_ASSERT( pFname );

    pFile = UT_LoadFileToMainMem( pFname );
    if( pFile != NULL )
    {
        if( NNS_G2dGetUnpackedCharacterData( pFile, ppCharData ) )
        {
            NNS_G2dPrintCharacterData( *ppCharData );
            return pFile;
        }
    
        UT_Free(pFile);
    }

    return NULL;
}

/*---------------------------------------------------------------------------*
  Name:         UT_SetUpNitroCharacterData3D

  Description:  Loads palette data, character data, cell data and animation data.
                Loads to VRAM as needed.

                          
  Arguments:    objData       : Address of the OBJ related structure
                nitroCharData : NITRO-CHARACTER file name structure
                texBaseAddr   : Texture base address
                plttBaseAddr  : Palette base address

  Returns:      None
 *---------------------------------------------------------------------------*/

void 
UT_SetUpNitroCharacterData3D(
    ObjData* objData, 
    const NitroCharacterData* nitroCharData,
    u32 texBaseAddr,
    u32 plttBaseAddr
)
{
    // Initialize Proxy
    NNS_G2dInitImageProxy( &objData->imageProxy );
    NNS_G2dInitImagePaletteProxy( &objData->paletteProxy );
        
    //------------------------------------------------------------------------------
    // load palette data 
    if (nitroCharData->nameNCLR != NULL)
    {
        void* pBuf;
        NNSG2dPaletteData* pPlttData;
    
        pBuf = UT_LoadNCLR( &pPlttData, nitroCharData->nameNCLR );
        SDK_NULL_ASSERT( pBuf );
    
        DC_FlushRange( (void*)pPlttData->pRawData, pPlttData->szByte );
        NNS_G2dLoadPalette(
            pPlttData,
            plttBaseAddr,
            NNS_G2D_VRAM_TYPE_3DMAIN,
            &objData->paletteProxy);
    
        // Copied the palette data to VRAM, so release this pBuf.
        UT_Free( pBuf ); 
    }
        
    //------------------------------------------------------------------------------
    // load character data for 3D (software sprite)

    if (nitroCharData->nameNCBR != NULL)
    {
        objData->pFileChar = UT_LoadNCGR( &objData->pCharData, nitroCharData->nameNCBR );
        SDK_NULL_ASSERT( objData->pFileChar );
        
        DC_FlushRange( (void*)objData->pCharData->pRawData, objData->pCharData->szByte );  

        // VRAM transfer animation
        if (NNS_G2D_VRAMTRANSFERDATA_FLAG_MASK & 
            ( objData->pCharData->characterFmt >> NNS_G2D_VRAMTRANSFERDATA_FLAG_SHIFT ))
        {
            // Unlike other NNS_G2dLoad* functions, NNS_G2dLoadImageVramTransfer 
            // does not load character data into VRAM.
            // It only sets the attribute of character data to ImageProxy and HW.
            // Therefore, character data needs to be kept in main memory.
        
            NNS_G2dLoadImageVramTransfer(
                objData->pCharData,
                texBaseAddr,
                NNS_G2D_VRAM_TYPE_3DMAIN,
                &objData->imageProxy );
        }
        else
        {
            if (objData->pCharData->mapingType == GX_OBJVRAMMODE_CHAR_2D)
            {
                NNS_G2dLoadImage2DMapping( 
                    objData->pCharData, 
                    texBaseAddr, 
                    NNS_G2D_VRAM_TYPE_3DMAIN, 
                    &objData->imageProxy);  
            }
            else
            {
                NNS_G2dLoadImage1DMapping( 
                    objData->pCharData, 
                    texBaseAddr, 
                    NNS_G2D_VRAM_TYPE_3DMAIN, 
                    &objData->imageProxy);  
            }
    
            // Release the character data
            // Cell data and cell animation data are not released 
            // because they are used in main memory.
            UT_Free( objData->pFileChar );      
        }
    }
        
    //------------------------------------------------------------------------------
    // load cell data
    if (nitroCharData->nameNCER != NULL)
    {
        objData->pFileCell = UT_LoadNCER( &objData->pCellBank, nitroCharData->nameNCER );
        SDK_NULL_ASSERT( objData->pFileCell );
    }

    //------------------------------------------------------------------------------
    // load animation data
    if (nitroCharData->nameNANR != NULL)
    {
        objData->pFileAnim = UT_LoadNANR( &objData->pAnimBank, nitroCharData->nameNANR );
        SDK_NULL_ASSERT( objData->pFileAnim );
    }   
}

/*---------------------------------------------------------------------------*
  Name:         UT_SetUpNitroCharacterData2D

  Description:  Loads palette data, character data, cell data and animation data.
                Loads to VRAM as needed.
                Note that this is not versatile.

                          
  Arguments:    objData       : Address of the OBJ related structure
                nitroCharData : NITRO-CHARACTER file name structure
                texBaseAddr   : Texture base address
                plttBaseAddr  : Palette base address

  Returns:      None
 *---------------------------------------------------------------------------*/

void 
UT_SetUpNitroCharacterData2D(
    ObjData* objData, 
    const NitroCharacterData* nitroCharData,
    u32 charBaseAddr,
    u32 plttBaseAddr
)
{
    void* pBuf;
    NNSG2dPaletteData* pPlttData;

    // Initialize Proxy
    NNS_G2dInitImageProxy( &objData->imageProxy );
    NNS_G2dInitImagePaletteProxy( &objData->paletteProxy );

    // Loads NCLR data to main memory.
    pBuf = UT_LoadNCLR( &pPlttData, nitroCharData->nameNCLR );
    SDK_NULL_ASSERT( pBuf );
    DC_FlushRange( (void*)pPlttData->pRawData, pPlttData->szByte );

    // Loads palette data to VRAM.

    NNS_G2dLoadPalette(
            pPlttData,
            plttBaseAddr,
            NNS_G2D_VRAM_TYPE_2DMAIN,
            &objData->paletteProxy);

    NNS_G2dLoadPalette(
            pPlttData,
            plttBaseAddr,
            NNS_G2D_VRAM_TYPE_2DSUB, 
            &objData->paletteProxy);

    // Copied the palette data to VRAM, so release this pBuf.
    UT_Free( pBuf );

    // Loads NCGR data to main memory.
    pBuf = UT_LoadNCGR( &objData->pCharData, nitroCharData->nameNCGR );
    SDK_NULL_ASSERT( objData->pCharData );    
    DC_FlushRange( (void*)objData->pCharData->pRawData, objData->pCharData->szByte );  

    //  Copies character data to VRAM.

    NNS_G2dLoadImage1DMapping( 
                    objData->pCharData, 
                    charBaseAddr, 
                    NNS_G2D_VRAM_TYPE_2DMAIN, 
                    &objData->imageProxy);  

    NNS_G2dLoadImage1DMapping( 
                    objData->pCharData, 
                    charBaseAddr, 
                    NNS_G2D_VRAM_TYPE_2DSUB, 
                    &objData->imageProxy);  

    // Copied the character data to VRAM, so release this pBuf.
    UT_Free( pBuf );      

    // Loads NCER data to main memory.
    objData->pFileCell = UT_LoadNCER( &objData->pCellBank, nitroCharData->nameNCER );
    // Cell data is not released because they are used in main memory.
}


