/*---------------------------------------------------------------------------*
  Project:  G2D Program Sample
  File:     yoshi_common.c

  Copyright 2004 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Log: $
  $NoKeywords: $
 *---------------------------------------------------------------------------*/
 
#include <nitro.h>
#include <nnsys/g2d.h>
#include "vram_map.h"
#include "geography.h"
#include "utility.h"
#include "task_priority.h"
#include "yoshi.h"
#include "effect.h"
#include "renderer.h"

const NNSG2dImageAttr yoshiTexAttr = 
{ 
    GX_TEXSIZE_S128,        // GXTexSizeS
    GX_TEXSIZE_T256,        // GXTexSizeT
    GX_TEXFMT_PLTT256,      // GXTextFmt
    FALSE,                  // bExtendedPlt
    GX_TEXPLTTCOLOR0_TRNS,  // plttUse
    GX_OBJVRAMMODE_CHAR_2D  // mappingType
};

/*---------------------------------------------------------------------------*
  Name:         MoveX_WithGroundCheckYoshi

  Description:  Moves to X direction and determines the hit to the landscape
                        
  Arguments:    None
                
  Returns:      None
 *---------------------------------------------------------------------------*/

void MoveX_WithGroundCheckYoshi(YoshiTask* task)
{
    fx32 centerHeight;
    fx32 pos_x_backup = task->pos.x;                 // Save the current X-coordinate
    fx32 centerDiff;
    
    task->pos.x += task->velocity.x;                 // X movement
    centerHeight = GetGeographicHeight(task->pos.x); // Obtains the height of the field
    centerDiff = FX_ABS(centerHeight - task->pos.y); // Vertical difference between the field and character

    if (centerDiff < FX32_ONE * 2)                   // If the vertical difference is less than 2, align the level.
    {                                                // Effect for going up a hill.
        task->pos.y = centerHeight;
    }
    else if (centerHeight < task->pos.y)             // If the ground level was high, cancel the horizontal movement.
    {                                                // Effect for crashing into the wall
        task->pos.x = pos_x_backup;
    }
    else if (( task->common.func != YoshiJumpUp)
            &&(task->common.func != YoshiJumpDown)
            &&(task->common.func != YoshiJumpUp_Init)
            &&(task->common.func != YoshiAirWalk)
            &&(task->common.func != YoshiAirWalk_Init)
            ) // If Yoshi is above the ground, downward movement                                    
    {
        task->velocity.y = 0;
        task->common.func = YoshiJumpDown_Init;            
    }
    
    gYoshiPos.x = task->pos.x;    // Sets the Yoshi coordinates to the global variable
    gYoshiPos.y = task->pos.y;    // Sets the Yoshi coordinates to the global variable
}

/*---------------------------------------------------------------------------*
  Name:         MoveY_WithGroundCheckYoshi

  Description:  Moves to Y direction and determines the hit to the landscape
                        
  Arguments:    None
                
  Returns:      None
 *---------------------------------------------------------------------------*/

void MoveY_WithGroundCheckYoshi(YoshiTask* task)
{
    fx32 centerHeight = GetGeographicHeight(task->pos.x);   // Obtains the height of the field

    task->pos.y -= task->velocity.y;            // Y movement
    if (task->velocity.y > FX32_ONE * -4)       // -4 = infall velocity limit
    {
        task->velocity.y -= FX32_ONE * 0.2;     // Increasing the infall velocity
    }
    
    if (task->common.func == YoshiJumpUp)
    {
        if (task->velocity.y <= 0)              // End of the upward movement, then downwards.
        {
            task->velocity.y = 0;
            task->common.func = YoshiJumpDown_Init;
        }   
    }
    else if (task->common.func == YoshiJumpDown || task->common.func == YoshiAirWalk)
    {               
        if (task->pos.y > centerHeight)         // Land on the ground
        {
            task->pos.y = centerHeight; 
            task->common.func = YoshiBrake_Init;
            
            // Generates the clouds of dust task to coordinate with landing
            {
                EffectTask* effect;
                effect = (EffectTask *)TaskCtrl_SetTask(Smoke_Init, TASK_PRIORITY_EFFECT_NOTRENDER_ALPHA);
                effect->pos = task->pos;                // Sets the location of clouds of dust
                effect->velocity = task->velocity;      // Sets the velocity of clouds of dust
            }
        }       
    }
    
    gYoshiPos.x = task->pos.x;    // Sets the Yoshi coordinates to the global variable
    gYoshiPos.y = task->pos.y;    // Sets the Yoshi coordinates to the global variable
}

/*---------------------------------------------------------------------------*
  Name:         DrawYoshi

  Description:  Yoshi common drawing function
                        
  Arguments:    None
                
  Returns:      None
 *---------------------------------------------------------------------------*/

void DrawYoshi(void)
{
    YoshiTask* pSelfTask = (YoshiTask *)TaskCtrl_GetCurrentTask();
    u16 rotate  = 0;                            // Rotation angle
    NNSG2dFVec2  scale = {FX32_ONE, FX32_ONE};  // Magnification 
    NNSG2dFVec2  pos;                           // Cell drawing position

    if (pSelfTask->direction == RIGHT)
    {
        scale.x *= -1;                          // Flip substitution
    }

    // Display X-coordinate calculation
    pos.x = FX32_SCREEN_CENTER_X;

    // Setting the H scroll value of BG
    // Obtain the X-coordinate of left edge of the screen
    gBgHScroll = (pSelfTask->pos.x - FX32_SCREEN_CENTER_X) >> FX32_SHIFT;

    // Display Y-coordinate calculation
    if (pSelfTask->pos.y > YOSHI_UPDOWN_BORDER) // With YOSHI_UPDOWN_BORDER or more, the position of Yoshi moves up and down.
    {
        gBgVScroll = 0;                         // Setting the V scroll value of BG
        pos.y = pSelfTask->pos.y - YOSHI_HEIGHT_HALF;
    }
    else if (pSelfTask->pos.y < 0)              // Also with 0 or smaller, Yoshi goes up and down
    {
        gBgVScroll =  (-YOSHI_UPDOWN_BORDER) >> FX32_SHIFT;   // Setting the V scroll value of BG
        pos.y = YOSHI_UPDOWN_BORDER - YOSHI_HEIGHT_HALF + pSelfTask->pos.y;               
    }
    else                                        // With YOSHI_UPDOWN_BORDER or lower, it places at center, and moves the BG.
    {
        gBgVScroll = (pSelfTask->pos.y - YOSHI_UPDOWN_BORDER) >> FX32_SHIFT;   // Setting the V scroll value of BG0
        pos.y = YOSHI_UPDOWN_BORDER - YOSHI_HEIGHT_HALF;
    }

    // Blink display when damaged
    if (pSelfTask->injuryTime > 0)
    {
        pSelfTask->injuryTime--;
    }   
    if (pSelfTask->injuryTime & 8)
    {
        return;
    }

    // Material setting
    MaterialSetup();

    // Draw
    //
    // Draw cells by using Renderer module.
    //    
    {
        // Sets the image data and palette data referenced by the renderer.
          NNS_G2dSetRendererImageProxy( &render3D, &objDataYoshi.imageProxy, &objDataYoshi.paletteProxy );  
            NNS_G2dPushMtx();
                NNS_G2dTranslate( pos.x, pos.y, 0 );
                NNS_G2dRotZ( FX_SinIdx( rotate ), FX_CosIdx( rotate ) );
                NNS_G2dScale( scale.x, scale.y, FX32_ONE );
                NNS_G2dDrawCellAnimation( pSelfTask->cellAnim );
            NNS_G2dPopMtx();
    }       
}
